__version__ = "1.5.1"

__title__ = "glassesTools"
__description__ = "Tools for processing wearable eye tracker recordings."
__url__ = "https://github.com/dcnieho/glassesTools"
__uri__ = __url__
__doc__ = __description__ + " <" + __url__ + ">"

__author__ = "Diederick Niehorster"
__email__ = "diederick_c.niehorster@humlab.lu.se"

__license__ = "MIT"
__copyright__ = "Copyright (c) 2022-2024 " + __author__
